# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from ._numericimpl import NumericOperationsImpl, NullableNumericOperationsImpl
from ._stringimpl import StringOperationsImpl, NullableStringOperationsImpl
from ._boolimpl import BooleanOperationsImpl, NullableBooleanOperationsImpl
from ._interface import OperationsBlock
from ._shapeimpl import UniformShapeOperations

__all__ = [
    "NumericOperationsImpl",
    "StringOperationsImpl",
    "BooleanOperationsImpl",
    "UniformShapeOperations",
    "NullableNumericOperationsImpl",
    "NullableStringOperationsImpl",
    "NullableBooleanOperationsImpl",
    "OperationsBlock",
    "binary_op",
    "unary_op",
    "variadic_op",
    "split_nulls_and_values",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from typing import TYPE_CHECKING, Union

import numpy as np
import spox.opset.ai.onnx.v20 as op
from spox import Var
from typing_extensions import Self

import ndonnx as ndx

from ._index import ScalarIndexType, construct_index
from ._utility import get_dtype, get_rank, get_shape

if TYPE_CHECKING:
    from ._data_types import CoreType

IndexType = Union[ScalarIndexType, tuple[ScalarIndexType, ...], "_CoreArray"]


class _CoreArray:
    """Thin wrapper around a `spox.Var` as well as optionally the eager propagated
    value."""

    var: Var
    _eager_value: np.ndarray | None

    def __init__(self, value: Var | np.ndarray) -> None:
        if isinstance(value, Var):
            self.var = value
            self._eager_value = None
        else:
            self.var = op.const(value)
            self._eager_value = value

    def _set(self, other: _CoreArray) -> Self:
        self.var = other.var
        self._eager_value = other.to_numpy()
        return self

    def copy(self) -> Self:
        return type(self)(
            self.var if self._eager_value is None else self._eager_value,
        )

    def to_numpy(self) -> np.ndarray | None:
        return self._eager_value

    @property
    def dtype(self) -> CoreType:
        return ndx.from_numpy_dtype(get_dtype(self))

    def astype(self, dtype: CoreType) -> _CoreArray:
        import ndonnx._opset_extensions as opx

        return opx.cast(self, dtype)

    def __getitem__(self, index: IndexType) -> _CoreArray:
        import ndonnx._opset_extensions as opx

        normalised_index = self._normalise_index(index)
        return opx.getitem(self, normalised_index)

    def __setitem__(
        self, index: IndexType, updates: int | bool | float | _CoreArray
    ) -> Self:
        import ndonnx._opset_extensions as opx

        normalised_index = self._normalise_index(index)
        normalised_updates = (
            updates
            if isinstance(updates, _CoreArray)
            else _CoreArray(np.array(updates))
        )
        ret = opx.setitem(self, normalised_index, normalised_updates)
        return self._set(ret)

    def _normalise_index(
        self, index: IndexType
    ) -> _CoreArray | tuple[ScalarIndexType, ...]:
        if isinstance(index, _CoreArray):
            return index
        else:
            index = construct_index(self, index)
            indexing_expressions = len(tuple(idx for idx in index if idx is not None))
            if self.ndim != indexing_expressions:
                raise IndexError(
                    f"Index has {indexing_expressions} expressions but Array has rank {self.ndim}"
                )
            return index

    @property
    def ndim(self) -> int:
        return get_rank(self)

    def __repr__(self) -> str:
        return f"<CoreArray : {self.dtype}[{get_shape(self)}] -> propagated: {self.to_numpy()}>"



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, TypeAlias, TypeVar

import numpy as np

import ndonnx as ndx
from ndonnx import _opset_extensions as opx

if TYPE_CHECKING:
    from ndonnx import Array

Scalar = TypeVar("Scalar", int, float, str)

Key: TypeAlias = Scalar
Value = TypeVar("Value", int, float, str)


def shape(x: Array) -> Array:
    """Returns shape of an array.

    Parameters
    ----------
    x: Array
        Array to get shape of

    Returns
    -------
    out: Array
        Array of shape
    """
    out = x.dtype._ops.shape(x)
    if out is NotImplemented:
        raise ndx.UnsupportedOperationError(f"`shape` not implemented for `{x.dtype}`")
    return out


def isin(x: Array, items: Sequence[Scalar]) -> Array:
    """Return true where the input ``Array`` contains an element in ``items``.

    Parameters
    ----------
    x: Array
        The input Array to check for the presence of items.
    items: Sequence[Scalar]
        Scalar items to check for in the input Array.

    Returns
    -------
    out: Array
        Array of booleans indicating whether each element of ``x`` is in ``items``.
    """
    if isinstance(x.dtype, ndx.Nullable):
        return isin(x.values, items) & ~x.null
    elif len(items) == 1:
        if isinstance(items[0], float) and np.isnan(items[0]):
            return ndx.isnan(x)
        else:
            return x == items[0]
    else:
        mapping = dict(zip(items, (1,) * len(items)))
        return static_map(x, mapping, 0).astype(ndx.bool)


def static_map(
    x: Array,
    mapping: Mapping[Key, Value],
    default: Value | None = None,
) -> Array:
    """Map values in ``x`` based on the static ``mapping``.

    Parameters
    ----------
    x: Array
        The Array whose values will be mapped.
    mapping: Mapping[Key, Value]
        A mapping from keys to values. The keys must be of the same type as the values in ``x``.
    default: Value, optional
        The default value to use when a key is not found in the mapping.

    Returns
    -------
    out: Array
        A new Array with the values mapped according to the mapping.
    """
    if not isinstance(x.dtype, ndx.CoreType):
        raise TypeError("static_map accepts only non-nullable arrays")
    data = opx.static_map(x._core(), mapping, default)
    return ndx.Array._from_fields(data.dtype, data=data)


def fill_null(x: Array, value: Array | Scalar) -> Array:
    """Returns a new ``Array`` with the null values filled with the given value.

    Parameters
    ----------
    x: Array
        The Array to fill the null values of.
    value: Array | Scalar
        The value to fill the null values with.

    Returns
    -------
    out: Array
        A new Array with the null values filled with the given value.
    """

    out = x.dtype._ops.fill_null(x, value)
    if out is NotImplemented:
        raise ndx.UnsupportedOperationError(
            f"`fill_null` not implemented for `{x.dtype}`"
        )
    return out


def make_nullable(x: Array, null: Array) -> Array:
    """Given an array ``x`` of values and a null mask ``null``, construct a new Array
    with a nullable data type.

    Parameters
    ----------
    x: Array
        Array of values

    null: Array
        Array of booleans indicating whether each element of ``x`` is null.

    Returns
    -------
    out: Array
        A new Array with a nullable data type.

    Raises
    ------
    TypeError
        If the data type of ``x`` does not have a nullable counterpart.
    """
    out = x.dtype._ops.make_nullable(x, null)
    if out is NotImplemented:
        raise ndx.UnsupportedOperationError(
            f"'make_nullable' not implemented for `{x.dtype}`"
        )
    return out


def _static_shape(x: Array) -> tuple[int | None, ...]:
    """Return shape of the array as a tuple. Typical implementations will make use of
    ONNX shape inference, with `None` entries denoting unknown or symbolic dimensions.

    Parameters
    ----------
    x: Array
        Array to get static shape of

    Returns
    -------
    out: tuple[int | None, ...]
    """
    out = x.dtype._ops.static_shape(x)
    if out is NotImplemented:
        raise ndx.UnsupportedOperationError(
            f"`static_shape` not implemented for `{x.dtype}`"
        )
    return out



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from abc import abstractmethod

import numpy as np
import typing_extensions

from ndonnx._core._interface import OperationsBlock
from ndonnx._data_types.coretype import CoreType

from .schema import Schema


class StructType:
    """The ``StructType`` class defines all data types that go beyond the core data
    types present in the ONNX standard.

    Inheriting from ``StructType`` enables the implementation of more complex and domain-specific data types. The API for doing so is used to implement the nullable data types exported by ``ndonnx``. This API is not yet public but may be considered ready for experimentation.
    """

    @abstractmethod
    def _fields(self) -> dict[str, StructType | CoreType]:
        """Define the named fields of this StructType."""
        ...

    @abstractmethod
    def _parse_input(self, x: np.ndarray) -> dict:
        """This method may be used by runtime utilities. It should break down the input
        numpy array into a dictionary of values with an entry for each field in this
        StructType.

        As an example, consider defining a 128 bit data type as a StructType as follows:

            .. code-block:: python

                class My128BitType(StructType):
                    def __init__(self):
                        super().__init__(low=uint64(), high=uint64())

            Then, ``parse_input`` can be defined as follows:

            .. code-block:: python

                def _parse_input(self, x: np.ndarray) -> dict[str, np.ndarray]:
                    # We expect a numpy array of python int objects.
                    if x.dtype != object:
                        raise TypeError("Input must be an object array")
                    mask = (1 << 64) - 1
                    return {
                        "low": (x & mask).astype(np.uint64),
                        "high": (x >> 64).astype(np.uint64) & mask,
                    }
        """
        ...

    @abstractmethod
    def _assemble_output(self, fields: dict[str, np.ndarray]) -> np.ndarray:
        """This method may be used by runtime utilities, and is essentially an inverse
        of ``parse_input``. It should take a dictionary of values (representing the
        fields of the struct being defined) and compose them into a single
        representation of a value with this data type.

        As an example, consider defining a 128 bit data type as a StructType as follows:

        .. code-block:: python

            class My128BitType(StructType):
                def __init__(self):
                    super().__init__(low=uint64(), high=uint64())

        Then, ``assemble_output`` can be defined as follows:

        .. code-block:: python

            def _assemble_output(self, fields: dict[str, np.ndarray]) -> np.ndarray:
                return (fields["high"].astype(object) << 64) | fields["low"].astype(object)
        """

    @abstractmethod
    def _schema(self) -> Schema:
        """Exports a schema that describes any basic information required by runtime
        utilities to consume input values corresponding to this array."""
        ...

    @abstractmethod
    def copy(self) -> typing_extensions.Self:
        """Clone this data type."""
        ...

    def __eq__(self, other: object) -> bool:
        return type(self) is type(other) or type(self) is other

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return type(self).__name__

    def __hash__(self) -> int:
        return hash(type(self))

    _ops: OperationsBlock
    """Defines functions that may be executed using a ``ndx.Array`` of this data
    type."""



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import builtins
from collections import namedtuple
from typing import Literal

import numpy as np
import numpy.typing as npt

import ndonnx._data_types as dtypes
from ndonnx._data_types import CastError, CastMixin, CoreType, _NullableCore
from ndonnx._data_types.structtype import StructType

from . import _opset_extensions as opx
from ._array import Array, _from_corearray
from ._utility import (
    promote,
)


class UnsupportedOperationError(TypeError): ...


# creation.py


def arange(
    start,
    stop=None,
    step=1,
    *,
    dtype: dtypes.CoreType | dtypes.StructType | None = None,
    device=None,
):
    if dtype is None:
        if builtins.all(
            isinstance(asarray(x).dtype, dtypes.Integral)
            for x in [start, stop, step]
            if x is not None
        ):
            dtype = dtypes.int64
        else:
            dtype = dtypes.float64

    out = dtype._ops.arange(start, stop, step, dtype=dtype, device=device)
    if out is NotImplemented:
        raise UnsupportedOperationError(
            f"Unsupported operand type for arange: '{dtype}'"
        )
    return out


def asarray(
    x: npt.ArrayLike | Array,
    dtype: dtypes.CoreType | dtypes.StructType | None = None,
    copy=None,
    device=None,
) -> Array:
    if not isinstance(x, Array):
        arr = np.asanyarray(
            x,
            dtype=(
                dtype.to_numpy_dtype() if isinstance(dtype, dtypes.CoreType) else None
            ),
        )
        if dtype is None:
            dtype = dtypes.from_numpy_dtype(arr.dtype)
            if isinstance(arr, np.ma.masked_array):
                dtype = dtypes.into_nullable(dtype)

        ret = Array._construct(
            shape=arr.shape, dtype=dtype, eager_values=dtype._parse_input(arr)
        )
    else:
        ret = x.copy() if copy is True else x

    if dtype is not None:
        ret = ret.astype(dtype)

    return ret


def empty(shape, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None):
    dtype = dtype or dtypes.float64
    if (out := dtype._ops.empty(shape, dtype=dtype)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for empty: '{dtype}'")


def empty_like(
    x, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None
):
    dtype = dtype or x.dtype
    if (out := dtype._ops.empty_like(x, dtype=dtype)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for empty_like: '{dtype}'"
    )


def eye(
    n_rows,
    n_cols=None,
    k=0,
    dtype: dtypes.StructType | dtypes.CoreType | None = None,
    device=None,
):
    dtype = dtype or dtypes.float64
    out = dtype._ops.eye(n_rows, n_cols=n_cols, k=k, dtype=dtype, device=device)
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for eye: '{dtype}'")


def full(
    shape,
    fill_value,
    *,
    dtype: dtypes.CoreType | dtypes.StructType | None = None,
    device=None,
):
    dtype = asarray(fill_value).dtype if dtype is None else dtype
    if (
        out := dtype._ops.full(shape, fill_value, dtype=dtype, device=device)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for full: '{dtype}'")


def full_like(
    x,
    fill_value,
    dtype: dtypes.CoreType | dtypes.StructType | None = None,
    device=None,
):
    if (
        out := x.dtype._ops.full_like(x, fill_value, dtype=dtype, device=device)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for full_like: `{x.dtype}`"
    )


def linspace(
    start: int | float,
    stop: int | float,
    num: int,
    *,
    dtype: dtypes.CoreType | dtypes.StructType | None = None,
    device=None,
    endpoint: bool = True,
) -> Array:
    dtype = dtypes.float64 if dtype is None else dtype
    out = dtype._ops.linspace(
        start, stop, num=num, dtype=dtype, endpoint=endpoint, device=device
    )
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for linspace: '{dtype}'")


def ones(shape, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None):
    dtype = dtypes.float64 if dtype is None else dtype
    if (out := dtype._ops.ones(shape, dtype=dtype)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for ones: '{dtype}'")


def ones_like(x, dtype: dtypes.StructType | dtypes.CoreType | None = None, device=None):
    if (out := x.dtype._ops.ones_like(x, dtype=dtype)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for ones_like: '{x.dtype}'"
    )


def tril(x: Array, *, k: int = 0) -> Array:
    out = x.dtype._ops.tril(x, k=k)
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for tril: '{x.dtype}'")


def triu(x: Array, *, k: int = 0) -> Array:
    out = x.dtype._ops.triu(x, k=k)
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for triu: '{x.dtype}'")


def zeros(shape, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None):
    dtype = dtypes.float64 if dtype is None else dtype
    if (out := dtype._ops.zeros(shape, dtype=dtype)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for zeros: '{dtype}'")


def zeros_like(
    x, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None
):
    if (out := x.dtype._ops.zeros_like(x, dtype=dtype)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for zeros_like: '{x.dtype}'"
    )


# data_type.py


def astype(x: Array, dtype, copy=True):
    if isinstance(dtype, np.dtype):
        dtype = dtypes.from_numpy_dtype(dtype)
    if x.dtype == dtype:
        return x if not copy else x.copy()
    if isinstance(x.dtype, CastMixin):
        try:
            out = x.dtype._cast_to(x, dtype)
            if copy:
                return out
            else:
                x._set(out)
                return x
        except CastError:
            pass
    if isinstance(dtype, CastMixin):
        try:
            out = dtype._cast_from(x)
            if copy:
                return out
            else:
                x._set(out)
                return x
        except CastError:
            pass
    raise CastError(f"Cannot cast {x.dtype} to {dtype}")


def can_cast(from_, to):
    if (out := from_._ops.can_cast(from_, to)) is not NotImplemented:
        return out
    elif (out := to._ops.can_cast(from_, to)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for can_cast: '{from_}' and '{to}'"
    )


def finfo(dtype):
    return dtypes.get_finfo(dtype)


def iinfo(dtype):
    return dtypes.get_iinfo(dtype)


def isdtype(dtype, kind) -> bool:
    if isinstance(kind, str):
        if kind == "bool":
            return dtype == dtypes.bool
        elif kind == "signed integer":
            return dtype in (dtypes.int8, dtypes.int16, dtypes.int32, dtypes.int64)
        elif kind == "unsigned integer":
            return dtype in (dtypes.uint8, dtypes.uint16, dtypes.uint32, dtypes.uint64)
        elif kind == "integral":
            return isinstance(dtype, dtypes.Integral)
        elif kind == "real floating":
            return isinstance(dtype, dtypes.Floating)
        elif kind == "complex floating":
            raise ValueError("'complex floating' is not supported")
        elif kind == "numeric":
            return isinstance(dtype, dtypes.Numerical)
    elif isinstance(kind, dtypes.CoreType):
        return dtype == kind
    elif isinstance(kind, tuple):
        return builtins.any(isdtype(dtype, k) for k in kind)
    raise TypeError(f"kind must be a string or a dtype, not {type(kind)}")


def result_type(
    *objects: dtypes.CoreType | dtypes.StructType | Array,
) -> dtypes.CoreType | dtypes.StructType:
    """Find the common data type for the provided objects."""
    observed_dtypes = {obj.dtype if isinstance(obj, Array) else obj for obj in objects}
    nullable = False
    np_dtypes = []
    for dtype in observed_dtypes:
        if isinstance(dtype, dtypes.StructType):
            if isinstance(dtype, _NullableCore):
                nullable = True
                np_dtypes.append(dtype.values.to_numpy_dtype())
            else:
                raise TypeError("result_type is not defined for arbitrary Struct types")
        else:
            np_dtypes.append(dtype.to_numpy_dtype())

    ret_dtype = dtypes.from_numpy_dtype(np.result_type(*np_dtypes))
    if nullable:
        return dtypes.into_nullable(ret_dtype)
    else:
        return ret_dtype


# elementwise.py


def abs(x):
    return _unary(x.dtype._ops.abs, x)


def acos(x):
    return _unary(x.dtype._ops.acos, x)


def acosh(x):
    return _unary(x.dtype._ops.acosh, x)


def add(x, y):
    return _binary("add", x, y)


def asin(x):
    return _unary(x.dtype._ops.asin, x)


def asinh(x):
    return _unary(x.dtype._ops.asinh, x)


def atan(x):
    return _unary(x.dtype._ops.atan, x)


def atan2(y, x):
    return _binary("atan2", y, x)


def atanh(x):
    return _unary(x.dtype._ops.atanh, x)


def bitwise_and(x, y):
    return _binary("bitwise_and", x, y)


def bitwise_left_shift(x, y):
    return _binary("bitwise_left_shift", x, y)


def bitwise_invert(x):
    return _unary(x.dtype._ops.bitwise_invert, x)


def bitwise_or(x, y):
    return _binary("bitwise_or", x, y)


def bitwise_right_shift(x, y):
    return _binary("bitwise_right_shift", x, y)


def bitwise_xor(x, y):
    return _binary("bitwise_xor", x, y)


def ceil(x):
    return _unary(x.dtype._ops.ceil, x)


def cos(x):
    return _unary(x.dtype._ops.cos, x)


def cosh(x):
    return _unary(x.dtype._ops.cosh, x)


def divide(x, y):
    return _binary("divide", x, y)


def equal(x, y):
    return _binary("equal", x, y)


def exp(x):
    return _unary(x.dtype._ops.exp, x)


def expm1(x):
    return _unary(x.dtype._ops.expm1, x)


def floor(x):
    return _unary(x.dtype._ops.floor, x)


def floor_divide(x, y):
    return _binary("floor_divide", x, y)


def greater(x, y):
    return _binary("greater", x, y)


def greater_equal(x, y):
    return _binary("greater_equal", x, y)


def isfinite(x):
    return logical_not(isinf(x))


def isinf(x):
    return _unary(x.dtype._ops.isinf, x)


def isnan(x):
    return _unary(x.dtype._ops.isnan, x)


def less(x, y):
    return _binary("less", x, y)


def less_equal(x, y):
    return _binary("less_equal", x, y)


def log(x):
    return _unary(x.dtype._ops.log, x)


def log1p(x):
    return _unary(x.dtype._ops.log1p, x)


def log2(x):
    return _unary(x.dtype._ops.log2, x)


def log10(x):
    return _unary(x.dtype._ops.log10, x)


def logaddexp(x, y):
    return _binary("logaddexp", x, y)


def logical_and(x, y):
    return _binary("logical_and", x, y)


def logical_not(x):
    return _unary(x.dtype._ops.logical_not, x)


def logical_or(x, y):
    return _binary("logical_or", x, y)


def logical_xor(x, y):
    return _binary("logical_xor", x, y)


def multiply(x, y):
    return _binary("multiply", x, y)


def negative(x):
    return _unary(x.dtype._ops.negative, x)


def not_equal(x, y):
    return _binary("not_equal", x, y)


def positive(x):
    return _unary(x.dtype._ops.positive, x)


def pow(x, y):
    return _binary("pow", x, y)


def remainder(x, y):
    return _binary("remainder", x, y)


def round(x):
    return _unary(x.dtype._ops.round, x)


def sign(x):
    return _unary(x.dtype._ops.sign, x)


def sin(x):
    return _unary(x.dtype._ops.sin, x)


def sinh(x):
    return _unary(x.dtype._ops.sinh, x)


def square(x):
    return multiply(x, x)


def sqrt(x):
    return _unary(x.dtype._ops.sqrt, x)


def subtract(x, y):
    return _binary("subtract", x, y)


def tan(x):
    return _unary(x.dtype._ops.tan, x)


def tanh(x):
    return _unary(x.dtype._ops.tanh, x)


def trunc(x):
    return _unary(x.dtype._ops.trunc, x)


# linalg.py


def matmul(x, y):
    return _binary("matmul", x, y)


def matrix_transpose(x):
    return _unary(x.dtype._ops.matrix_transpose, x)


# indexing.py


def take(x, indices, axis=None):
    if (out := x.dtype._ops.take(x, indices, axis)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for take: '{x.dtype}'")


# manipulation.py


# TODO: onnx shape inference doesn't work for 2 empty arrays
# TODO: what is the appropriate strategy to dispatch? (iterate over the inputs and keep trying is reasonable but it can
# change the outcome based on order if poorly implemented)


def broadcast_arrays(*arrays):
    if not arrays:
        return []
    it = iter(arrays)

    def numeric_like(x):
        if isinstance(x.dtype, (dtypes.NullableNumerical, dtypes.Numerical)):
            return x
        else:
            return zeros_like(x, dtype=dtypes.int64)

    ret = numeric_like(next(it))
    while (x := next(it, None)) is not None:
        ret = ret + numeric_like(x)
    target_shape = ret.shape
    return [broadcast_to(x, target_shape) for x in arrays]


def broadcast_to(x, shape):
    return x.dtype._ops.broadcast_to(x, shape)


# TODO: onnxruntime doesn't work for 2 empty arrays of integer type
# TODO: what is the appropriate strategy to dispatch? (iterate over the inputs and keep trying is reasonable but it can
# change the outcome based on order if poorly implemented)
def concat(arrays, axis=None):
    if axis is None:
        arrays = [reshape(x, [-1]) for x in arrays]
        axis = 0
    if builtins.any(not isinstance(array.dtype, CoreType) for array in arrays):
        raise UnsupportedOperationError(
            "concat is not supported for non-core types at this time"
        )
    arrays = promote(*arrays)
    return _from_corearray(opx.concat([array._core() for array in arrays], axis=axis))


def expand_dims(x, axis=0):
    return x.dtype._ops.expand_dims(x, axis)


def flip(x, axis=None):
    return x.dtype._ops.flip(x, axis=axis)


def permute_dims(x, axes):
    return x.dtype._ops.permute_dims(x, axes)


def reshape(x, shape, *, copy=None):
    return x.dtype._ops.reshape(x, shape, copy=copy)


def roll(x, shift, axis=None):
    return x.dtype._ops.roll(x, shift, axis)


def squeeze(x, axis):
    return x.dtype._ops.squeeze(x, axis)


def stack(arrays, axis=0):
    arrays = [expand_dims(x, axis=axis) for x in arrays]
    return concat(arrays, axis=axis)


# searching.py


def argmax(x, axis=None, keepdims=False):
    if (
        out := x.dtype._ops.argmax(x, axis=axis, keepdims=keepdims)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for argmax: '{x.dtype}'")


def argmin(x, axis=None, keepdims=False):
    if (
        out := x.dtype._ops.argmax(x, axis=axis, keepdims=keepdims)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for argmin: '{x.dtype}'")


def nonzero(x) -> tuple[Array, ...]:
    return _unary(x.dtype._ops.nonzero, x)


def searchsorted(
    x1, x2, *, side: Literal["left", "right"] = "left", sorter: Array | None = None
):
    if (
        out := x1.dtype._ops.searchsorted(x1, x2, side=side, sorter=sorter)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for searchsorted: '{x1.dtype}'"
    )


def where(condition, x, y):
    condition, x, y = map(asarray, (condition, x, y))
    if condition.dtype not in (dtypes.bool, dtypes.nbool):
        raise TypeError(
            f"condition must be a boolean array, received {condition.dtype}"
        )
    if (out := x.dtype._ops.where(condition, x, y)) is not NotImplemented:
        return out
    if (out := y.dtype._ops.where(condition, x, y)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported dtypes for where arguments: '{x.dtype}' and '{y.dtype}'"
    )


# set.py
def unique_all(x: Array):
    if (out := x.dtype._ops.unique_all(x)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for unique: '{x.dtype}'")


def unique_counts(x):
    ret = namedtuple("ret", ["values", "counts"])
    ret_all = unique_all(x)
    return ret(values=ret_all.values, counts=ret_all.counts)


def unique_inverse(x):
    ret = namedtuple("ret", ["values", "inverse_indices"])
    ret_all = unique_all(x)
    return ret(values=ret_all.values, inverse_indices=ret_all.inverse_indices)


def unique_values(x):
    return unique_all(x).values


# sorting.py


def argsort(x, *, axis=-1, descending=False, stable=True):
    if (
        out := x.dtype._ops.argsort(x, axis=axis, descending=descending, stable=stable)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for argsort: '{x.dtype}'"
    )


def sort(x, *, axis=-1, descending=False, stable=True):
    if (
        out := x.dtype._ops.sort(x, axis=axis, descending=descending, stable=stable)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for sort: '{x.dtype}'")


# statistical.py
def cumulative_sum(
    x,
    *,
    axis: int | None = None,
    dtype: CoreType | StructType | None = None,
    include_initial: bool = False,
):
    if (
        out := x.dtype._ops.cumulative_sum(
            x, axis=axis, dtype=dtype, include_initial=include_initial
        )
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for cumulative_sum: '{x.dtype}'"
    )


def max(x, *, axis=None, keepdims: bool = False):
    if (out := x.dtype._ops.max(x, axis=axis, keepdims=keepdims)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for max: '{x.dtype}'")


def mean(x, *, axis=None, keepdims: bool = False):
    if (
        out := x.dtype._ops.mean(x, axis=axis, keepdims=keepdims)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for max: '{x.dtype}'")


def min(x, *, axis=None, keepdims: bool = False):
    if (out := x.dtype._ops.min(x, axis=axis, keepdims=keepdims)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for max: '{x.dtype}'")


def prod(
    x, *, axis=None, dtype: CoreType | StructType | None = None, keepdims: bool = False
):
    if (
        out := x.dtype._ops.prod(x, dtype=dtype, axis=axis, keepdims=keepdims)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for prod: '{x.dtype}'")


def clip(x, /, min=None, max=None):
    if (out := x.dtype._ops.clip(x, min=min, max=max)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for clip: '{x.dtype}'")


def std(
    x,
    axis=None,
    keepdims: bool = False,
    correction=0.0,
    dtype: dtypes.StructType | dtypes.CoreType | None = None,
):
    dtype = x.dtype if dtype is None else dtype
    out = dtype._ops.std(
        x, axis=axis, keepdims=keepdims, correction=correction, dtype=dtype
    )
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for std: '{dtype}'")


def sum(
    x,
    *,
    axis=None,
    dtype: dtypes.StructType | dtypes.CoreType | None = None,
    keepdims: bool = False,
):
    if (
        out := x.dtype._ops.sum(x, axis=axis, dtype=dtype, keepdims=keepdims)
    ) is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for sum: '{x.dtype}'")


def var(
    x,
    *,
    axis=None,
    keepdims: bool = False,
    correction=0.0,
    dtype: dtypes.CoreType | dtypes.StructType | None = None,
):
    dtype = x.dtype if dtype is None else dtype
    out = dtype._ops.var(
        x, axis=axis, keepdims=keepdims, correction=correction, dtype=dtype
    )
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for var: '{dtype}'")


# utility.py


def all(x, *, axis=None, keepdims: bool = False):
    out = x.dtype._ops.all(x, axis=axis, keepdims=keepdims)
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for all: '{x.dtype}'")


def any(x, *, axis=None, keepdims: bool = False):
    out = x.dtype._ops.any(x, axis=axis, keepdims=keepdims)
    if out is not NotImplemented:
        return out
    raise UnsupportedOperationError(f"Unsupported operand type for any: '{x.dtype}'")


def _binary(func_name, x, y):
    x_dtype = asarray(x).dtype
    y_dtype = asarray(y).dtype
    if (out := getattr(x_dtype._ops, func_name)(x, y)) is not NotImplemented:
        return out
    if (out := getattr(y_dtype._ops, func_name)(x, y)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type(s) for {func_name}: '{x.dtype}' and '{y.dtype}'"
    )


def _unary(func, x):
    if (out := func(x)) is not NotImplemented:
        return out
    raise UnsupportedOperationError(
        f"Unsupported operand type for {func.__name__}: '{x.dtype}'"
    )


__all__ = [
    "arange",
    "asarray",
    "empty",
    "empty_like",
    "eye",
    "full",
    "full_like",
    "linspace",
    "ones",
    "ones_like",
    "tril",
    "triu",
    "zeros",
    "zeros_like",
    "astype",
    "broadcast_arrays",
    "broadcast_to",
    "can_cast",
    "finfo",
    "iinfo",
    "result_type",
    "abs",
    "acos",
    "acosh",
    "add",
    "asin",
    "asinh",
    "atan",
    "atan2",
    "atanh",
    "bitwise_and",
    "bitwise_left_shift",
    "bitwise_invert",
    "bitwise_or",
    "bitwise_right_shift",
    "bitwise_xor",
    "ceil",
    "cos",
    "cosh",
    "divide",
    "equal",
    "exp",
    "expm1",
    "floor",
    "floor_divide",
    "greater",
    "greater_equal",
    "isfinite",
    "isinf",
    "isnan",
    "less",
    "less_equal",
    "log",
    "log1p",
    "log2",
    "log10",
    "logaddexp",
    "logical_and",
    "logical_not",
    "logical_or",
    "logical_xor",
    "multiply",
    "negative",
    "not_equal",
    "positive",
    "pow",
    "remainder",
    "round",
    "sign",
    "sin",
    "sinh",
    "square",
    "sqrt",
    "subtract",
    "tan",
    "tanh",
    "trunc",
    "matmul",
    "matrix_transpose",
    "concat",
    "expand_dims",
    "flip",
    "permute_dims",
    "reshape",
    "roll",
    "squeeze",
    "stack",
    "argmax",
    "argmin",
    "nonzero",
    "searchsorted",
    "where",
    "unique_all",
    "unique_counts",
    "unique_inverse",
    "unique_values",
    "argsort",
    "sort",
    "cumulative_sum",
    "max",
    "mean",
    "min",
    "prod",
    "clip",
    "std",
    "sum",
    "var",
    "all",
    "any",
    "take",
]

for func_name in __all__:
    locals()[
        func_name
    ].__doc__ = f"``{func_name}`` is inherited from the Array API standard. Documentation can be found `here <https://data-apis.org/array-api/latest/API_specification/generated/array_api.{func_name}.html#{func_name}>`__."



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from ._additional import fill_null, make_nullable, isin, shape, static_map

__all__ = [
    "fill_null",
    "make_nullable",
    "isin",
    "shape",
    "static_map",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations
from ndonnx._utility import deprecated

from .aliases import (
    bool,
    float32,
    float64,
    int8,
    int16,
    int32,
    int64,
    nbool,
    nfloat32,
    nfloat64,
    nint8,
    nint16,
    nint32,
    nint64,
    nuint8,
    nuint16,
    nuint32,
    nuint64,
    nutf8,
    uint8,
    uint16,
    uint32,
    uint64,
    utf8,
)
from .classes import (
    Floating,
    Integral,
    Nullable,
    NullableFloating,
    NullableIntegral,
    NullableNumerical,
    NullableUnsigned,
    Numerical,
    Unsigned,
    _NullableCore,
    from_numpy_dtype,
    get_finfo,
    get_iinfo,
)
from .conversion import CastError, CastMixin
from .coretype import CoreType
from .schema import Schema
from .structtype import StructType


def into_nullable(dtype: StructType | CoreType) -> _NullableCore:
    """Return nullable counterpart, if present.

    Parameters
    ----------
    dtype : StructType | CoreType
        A data type exported by ``ndonnx``.

    Returns
    -------
    out : _NullableCore
        The nullable counterpart of the input type.

    Raises
    ------
    ValueError
        If the input type is unknown to ``ndonnx``.
    """
    if dtype == bool:
        return nbool
    elif dtype == float32:
        return nfloat32
    elif dtype == float64:
        return nfloat64
    elif dtype == int8:
        return nint8
    elif dtype == int16:
        return nint16
    elif dtype == int32:
        return nint32
    elif dtype == int64:
        return nint64
    elif dtype == uint8:
        return nuint8
    elif dtype == uint16:
        return nuint16
    elif dtype == uint32:
        return nuint32
    elif dtype == uint64:
        return nuint64
    elif dtype == utf8:
        return nutf8
    elif isinstance(dtype, _NullableCore):
        return dtype
    else:
        raise ValueError(f"Cannot promote {dtype} to nullable")


@deprecated(
    "Function 'ndonnx.promote_nullable' will be deprecated in ndonnx 0.7. "
    "To create nullable array, use 'ndonnx.additional.make_nullable' instead."
)
def promote_nullable(dtype: StructType | CoreType) -> _NullableCore:
    return into_nullable(dtype)


__all__ = [
    "CoreType",
    "StructType",
    "_NullableCore",
    "NullableFloating",
    "NullableIntegral",
    "NullableUnsigned",
    "NullableNumerical",
    "Nullable",
    "Floating",
    "Integral",
    "Unsigned",
    "Numerical",
    "bool",
    "from_numpy_dtype",
    "float32",
    "float64",
    "int16",
    "int32",
    "int64",
    "int8",
    "uint16",
    "uint32",
    "uint64",
    "uint8",
    "utf8",
    "nbool",
    "nfloat32",
    "nfloat64",
    "nint8",
    "nint16",
    "nint32",
    "nint64",
    "nuint8",
    "nuint16",
    "nuint32",
    "nuint64",
    "nutf8",
    "get_finfo",
    "get_iinfo",
    "promote_nullable",
    "Schema",
    "CastMixin",
    "CastError",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from ._core import OperationsBlock, UniformShapeOperations
from ._data_types import CastMixin, Schema, StructType

__all__ = [
    "StructType",
    "Schema",
    "CastMixin",
    "OperationsBlock",
    "UniformShapeOperations",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from .classes import (
    Boolean,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    NBoolean,
    NFloat32,
    NFloat64,
    NInt8,
    NInt16,
    NInt32,
    NInt64,
    NUInt8,
    NUInt16,
    NUInt32,
    NUInt64,
    NUtf8,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
    Utf8,
)

# Singleton exports of core data types
bool: Boolean = Boolean()
float32: Float32 = Float32()
float64: Float64 = Float64()
int16: Int16 = Int16()
int32: Int32 = Int32()
int64: Int64 = Int64()
int8: Int8 = Int8()
uint16: UInt16 = UInt16()
uint32: UInt32 = UInt32()
uint64: UInt64 = UInt64()
uint8: UInt8 = UInt8()
utf8: Utf8 = Utf8()

# Singleton exports of nullable instances of the core data types
nbool: NBoolean = NBoolean()
nfloat32: NFloat32 = NFloat32()
nfloat64: NFloat64 = NFloat64()
nint8: NInt8 = NInt8()
nint16: NInt16 = NInt16()
nint32: NInt32 = NInt32()
nint64: NInt64 = NInt64()
nuint8: NUInt8 = NUInt8()
nuint16: NUInt16 = NUInt16()
nuint32: NUInt32 = NUInt32()
nuint64: NUInt64 = NUInt64()
nutf8: NUtf8 = NUtf8()



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import builtins
import functools
import operator
import warnings
from collections import namedtuple
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Literal

import numpy as np

import ndonnx as ndx
import ndonnx._data_types as dtypes
import ndonnx._opset_extensions as opx
from ndonnx._utility import promote

from ._nullableimpl import NullableOperationsImpl
from ._shapeimpl import UniformShapeOperations
from ._utils import (
    binary_op,
    from_corearray,
    split_nulls_and_values,
    unary_op,
    validate_core,
    variadic_op,
    via_upcast,
)

if TYPE_CHECKING:
    from ndonnx import Array
    from ndonnx._corearray import _CoreArray


class NumericOperationsImpl(UniformShapeOperations):
    # elementwise.py

    @validate_core
    def abs(self, x):
        return unary_op(x, opx.abs)

    @validate_core
    def acos(self, x):
        return unary_op(x, opx.acos, dtypes.float32)

    @validate_core
    def acosh(self, x):
        return unary_op(x, opx.acosh, dtypes.float32)

    @validate_core
    def add(self, x, y) -> ndx.Array:
        return _via_i64_f64(opx.add, [x, y])

    @validate_core
    def asin(self, x):
        return unary_op(x, opx.asin, dtypes.float32)

    @validate_core
    def asinh(self, x):
        return unary_op(x, opx.asinh, dtypes.float32)

    @validate_core
    def atan(self, x):
        return unary_op(x, opx.atan, dtypes.float32)

    @validate_core
    def atan2(self, y, x):
        return self.atan(self.divide(y, x))

    @validate_core
    def atanh(self, x):
        return unary_op(x, opx.atanh, dtypes.float32)

    @validate_core
    def bitwise_and(self, x, y):
        return binary_op(x, y, opx.bitwise_and)

    # TODO: ONNX standard -> not cyclic
    @validate_core
    def bitwise_left_shift(self, x, y):
        return binary_op(
            x, y, lambda a, b: opx.bit_shift(a, b, direction="LEFT"), dtypes.uint64
        )

    @validate_core
    def bitwise_invert(self, x):
        return unary_op(x, opx.bitwise_not)

    @validate_core
    def bitwise_or(self, x, y):
        return binary_op(x, y, opx.bitwise_or)

    # TODO: ONNX standard -> not cyclic
    @validate_core
    def bitwise_right_shift(self, x, y):
        return binary_op(
            x,
            y,
            lambda x, y: opx.bit_shift(x, y, direction="RIGHT"),
            dtypes.uint64,
        )

    @validate_core
    def bitwise_xor(self, x, y):
        return binary_op(x, y, opx.bitwise_xor)

    @validate_core
    def ceil(self, x):
        if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
            return unary_op(x, opx.ceil, dtypes.float64)
        return ndx.asarray(x, copy=False)

    @validate_core
    def cos(self, x):
        return unary_op(x, opx.cos, dtypes.float32)

    @validate_core
    def cosh(self, x):
        return unary_op(x, opx.cosh, dtypes.float32)

    @validate_core
    def divide(self, x, y):
        x, y = promote(x, y)
        if not isinstance(x.dtype, (dtypes.Numerical, dtypes.NullableNumerical)):
            raise TypeError(f"Unsupported dtype for divide: {x.dtype}")
        bits = (
            ndx.iinfo(x.dtype).bits
            if isinstance(x.dtype, (dtypes.Integral, dtypes.NullableIntegral))
            else ndx.finfo(x.dtype).bits
        )
        via_dtype = (
            dtypes.float64
            if bits > 32 or x.dtype in (dtypes.nuint32, dtypes.uint32)
            else dtypes.float32
        )
        return variadic_op([x, y], opx.div, via_dtype=via_dtype, cast_return=False)

    @validate_core
    def equal(self, x, y) -> Array:
        x, y = promote(x, y)
        if isinstance(x.dtype, (dtypes.Integral, dtypes.NullableIntegral)):
            return variadic_op([x, y], opx.equal, dtypes.int64, cast_return=False)
        else:
            return binary_op(x, y, opx.equal)

    @validate_core
    def not_equal(self, x, y) -> Array:
        return ndx.logical_not(x == y)

    @validate_core
    def exp(self, x):
        return unary_op(x, opx.exp, dtypes.float32)

    @validate_core
    def expm1(self, x):
        return self.subtract(self.exp(x), 1)

    @validate_core
    def floor(self, x):
        x = ndx.asarray(x)
        if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
            return unary_op(x, opx.floor, dtypes.float64)
        return x

    @validate_core
    def floor_divide(self, x, y):
        x, y = promote(x, y)
        dtype = x.dtype
        out = self.floor(self.divide(x, y))
        if isinstance(
            dtype,
            (
                dtypes.Integral,
                dtypes.NullableIntegral,
                dtypes.Unsigned,
                dtypes.NullableUnsigned,
            ),
        ):
            out = out.astype(dtype)
        return out

    @validate_core
    def greater(self, x, y):
        return _via_i64_f64(opx.greater, [x, y], cast_return=False)

    @validate_core
    def greater_equal(self, x, y):
        return _via_i64_f64(opx.greater_or_equal, [x, y], cast_return=False)

    @validate_core
    def isfinite(self, x):
        return self.logical_not(self.isinf(x))

    @validate_core
    def isinf(self, x):
        if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
            return unary_op(x, opx.isinf)
        return ndx.full(x.shape, fill_value=False)

    @validate_core
    def isnan(self, x):
        if not isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
            return ndx.full_like(x, False, dtype=dtypes.bool)
        else:
            return unary_op(x, opx.isnan)

    @validate_core
    def less(self, x, y):
        return _via_i64_f64(opx.less, [x, y], cast_return=False)

    @validate_core
    def less_equal(self, x, y):
        return _via_i64_f64(opx.less_or_equal, [x, y], cast_return=False)

    @validate_core
    def log(self, x):
        return unary_op(x, opx.log, dtypes.float64)

    @validate_core
    def log1p(self, x):
        return self.add(self.log(x), ndx.asarray(1, x.dtype))

    @validate_core
    def log2(self, x):
        return ndx.log(x) / float(np.log(2))

    @validate_core
    def log10(self, x):
        return ndx.log(x) / float(np.log(10))

    @validate_core
    def logaddexp(self, x, y):
        return self.log(self.exp(x) + self.exp(y))

    @validate_core
    def multiply(self, x, y):
        x, y = promote(x, y)
        dtype = x.dtype
        via_dtype: dtypes.CoreType
        if isinstance(dtype, (dtypes.Integral, dtypes.NullableIntegral)) or dtype in (
            dtypes.nbool,
            dtypes.bool,
        ):
            via_dtype = dtypes.int64
        elif isinstance(dtype, (dtypes.Floating, dtypes.NullableFloating)):
            via_dtype = dtypes.float64
        else:
            raise TypeError(f"Unsupported dtype for multiply: {dtype}")
        return binary_op(x, y, opx.mul, via_dtype)

    @validate_core
    def negative(self, x):
        if isinstance(
            x.dtype, (dtypes.Unsigned, dtypes.NullableUnsigned)
        ) or x.dtype in (ndx.int16, ndx.nint16):
            return unary_op(x, opx.neg, dtypes.int64)
        return unary_op(x, opx.neg)

    @validate_core
    def positive(self, x):
        if not isinstance(x.dtype, (dtypes.Numerical, dtypes.NullableNumerical)):
            raise TypeError(f"Unsupported dtype for positive: {x.dtype}")
        return x.copy()

    @validate_core
    def pow(self, x, y):
        x, y = ndx.asarray(x), ndx.asarray(y)
        dtype = ndx.result_type(x, y)
        if isinstance(dtype, (dtypes.Unsigned, dtypes.NullableUnsigned)):
            return binary_op(x, y, opx.pow, dtypes.int64)
        else:
            return binary_op(x, y, opx.pow)

    @validate_core
    def remainder(self, x, y):
        return binary_op(x, y, lambda x, y: opx.mod(x, y, fmod=1))

    @validate_core
    def round(self, x):
        x = ndx.asarray(x)
        if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
            return unary_op(x, opx.round)
        else:
            return x

    @validate_core
    def sign(self, x):
        (x_values,), (x_null,) = split_nulls_and_values(x)
        out_values = ndx.where(x_values > 0, 1, ndx.where(x_values < 0, -1, 0)).astype(
            x_values.dtype
        )
        if x_null is None:
            return out_values
        else:
            return ndx.additional.make_nullable(out_values, x_null)

    @validate_core
    def sin(self, x):
        return unary_op(x, opx.sin, dtypes.float64)

    @validate_core
    def sinh(self, x):
        return via_upcast(opx.sinh, [x], float_dtype=dtypes.float32)

    @validate_core
    def square(self, x):
        return self.multiply(x, x)

    @validate_core
    def sqrt(self, x):
        return unary_op(x, opx.sqrt, dtypes.float32)

    @validate_core
    def subtract(self, x, y):
        x, y = promote(x, y)
        if isinstance(
            x.dtype, (dtypes.Unsigned, dtypes.NullableUnsigned)
        ) or x.dtype in (dtypes.int16, dtypes.int8, dtypes.nint16, dtypes.nint8):
            via_dtype = dtypes.int64
        else:
            via_dtype = None
        return binary_op(x, y, opx.sub, via_dtype=via_dtype)

    @validate_core
    def tan(self, x):
        return unary_op(x, opx.tan, dtypes.float32)

    @validate_core
    def tanh(self, x):
        return unary_op(x, opx.tanh, dtypes.float32)

    @validate_core
    def trunc(self, x):
        x = ndx.asarray(x)
        if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
            return ndx.where(x < 0, self.ceil(x), self.floor(x))
        return x

    # linalg.py

    @validate_core
    def matmul(self, x, y):
        return _via_i64_f64(opx.matmul, [x, y])

    @validate_core
    def matrix_transpose(self, x) -> ndx.Array:
        return ndx.permute_dims(x, list(range(x.ndim - 2)) + [x.ndim - 1, x.ndim - 2])

    # searching.py

    @validate_core
    def argmax(self, x, axis=None, keepdims=False):
        if axis is None:
            reshaped_x = ndx.reshape(x, [-1])._core()
            if keepdims:
                return from_corearray(
                    opx.reshape(
                        opx.arg_max(reshaped_x, axis=0, keepdims=False),
                        opx.const([1 for x in range(x.ndim)], dtype=dtypes.int64),
                    )
                )
            else:
                return from_corearray(
                    opx.reshape(
                        opx.arg_max(reshaped_x, axis=0, keepdims=False),
                        opx.const([], dtype=dtypes.int64),
                    )
                )
        return _via_i64_f64(lambda x: opx.arg_max(x, axis=axis, keepdims=keepdims), [x])

    @validate_core
    def argmin(self, x, axis=None, keepdims=False):
        if axis is None:
            reshaped_x = ndx.reshape(x, [-1])._core()
            if keepdims:
                return from_corearray(
                    opx.reshape(
                        opx.arg_min(reshaped_x, axis=0, keepdims=False),
                        opx.const([1 for x in range(x.ndim)], dtype=dtypes.int64),
                    )
                )
            else:
                return from_corearray(
                    opx.reshape(
                        opx.arg_min(reshaped_x, axis=0, keepdims=False),
                        opx.const([], dtype=dtypes.int64),
                    )
                )
        return _via_i64_f64(lambda x: opx.arg_min(x, axis=axis, keepdims=keepdims), [x])

    @validate_core
    def nonzero(self, x) -> tuple[Array, ...]:
        if not isinstance(x.dtype, dtypes.CoreType):
            return NotImplemented

        if x.ndim == 0:
            return (ndx.arange(0, x != 0, dtype=dtypes.int64),)

        ret_full_flattened = ndx.reshape(
            from_corearray(
                opx.ndindex(ndx.asarray(x.shape, dtype=dtypes.int64)._core())
            )[x != 0],
            [-1],
        )

        return tuple(
            ndx.reshape(
                from_corearray(
                    opx.gather_elements(
                        ret_full_flattened._core(),
                        ndx.arange(
                            i,
                            ret_full_flattened.shape[0],
                            x.ndim,
                            dtype=dtypes.int64,
                        )._core(),
                    )
                ),
                [-1],
            )
            for i in range(x.ndim)
        )

    @validate_core
    def searchsorted(
        self,
        x1,
        x2,
        *,
        side: Literal["left", "right"] = "left",
        sorter: ndx.Array | None = None,
    ):
        if x1.ndim != 1:
            raise TypeError("x1 must be a 1-dimensional array")
        if sorter is not None:
            if not isinstance(sorter.dtype, ndx.Integral):
                raise TypeError("sorter must be an array of integers")
            x1 = ndx.take(x1, sorter)

        if side not in ("left", "right"):
            raise ValueError("side must be 'left' or 'right'")

        nan_mask = ndx.isnan(x2)
        combined = ndx.concat([x1, x2])
        positions = ndx.unique_all(combined).inverse_indices + 1

        unique_x1 = ndx.unique_all(x1)
        counts = unique_x1.counts[unique_x1.inverse_indices]

        indices_x1, indices_x2 = map(
            from_corearray, opx.get_indices(x1._core(), x2._core(), positions._core())
        )

        how_many = ndx.zeros(ndx.asarray(combined.shape) + 1, dtype=dtypes.int64)
        how_many[
            ndx.where(indices_x1 + 1 <= combined.shape[0], indices_x1 + 1, indices_x1)
        ] = counts
        how_many = ndx.cumulative_sum(how_many, include_initial=True)

        ret = ndx.zeros(x2.shape, dtype=dtypes.int64)

        if side == "left":
            ret = how_many[indices_x2]
            ret[nan_mask] = ndx.asarray(x1.shape, dtype=dtypes.int64) - 1
        else:
            ret = how_many[indices_x2 + 1]
            ret[nan_mask] = ndx.asarray(x1.shape, dtype=dtypes.int64)

        return ret

    # set.py
    @validate_core
    def unique_all(self, x):
        new_dtype = x.dtype
        if isinstance(x.dtype, dtypes.Integral) or x.dtype in (
            dtypes.bool,
            dtypes.nbool,
        ):
            new_dtype = dtypes.int64
        elif isinstance(x.dtype, dtypes.Floating):
            warnings.warn(
                "Float unique_all is incomplete as nans are not handled correctly in onnxruntime"
            )
            new_dtype = dtypes.float64

        flattened = ndx.reshape(x, [-1])
        transformed = flattened.astype(new_dtype)._core()

        ret_opd = opx.unique(transformed, sorted=True)

        # FIXME: I think we can simply use arange/ones+cumsum or something for the indices
        # maybe: indices = opx.cumsum(ones_like(flattened, dtype=dtypes.i64), axis=ndx.asarray(0))
        indices = opx.squeeze(
            opx.ndindex(ndx.asarray(flattened.shape, dtype=dtypes.int64)._core()),
            opx.const([1], dtype=dtypes.int64),
        )

        ret = namedtuple("ret", ["values", "indices", "inverse_indices", "counts"])

        values = from_corearray(ret_opd[0])
        indices = from_corearray(indices[ret_opd[1]])
        inverse_indices = ndx.reshape(from_corearray(ret_opd[2]), x.shape)
        counts = from_corearray(ret_opd[3])

        return ret(
            values=values,
            indices=indices,
            inverse_indices=inverse_indices,
            counts=counts,
        )

    @validate_core
    def unique_counts(self, x):
        ret = namedtuple("ret", ["values", "counts"])
        ret_all = self.unique_all(x)
        return ret(values=ret_all.values, counts=ret_all.counts)

    @validate_core
    def unique_inverse(self, x):
        ret = namedtuple("ret", ["values", "inverse_indices"])
        ret_all = self.unique_all(x)
        return ret(values=ret_all.values, inverse_indices=ret_all.inverse_indices)

    @validate_core
    def unique_values(self, x):
        return self.unique_all(x).values

    # sorting.py

    @validate_core
    def argsort(self, x, *, axis=-1, descending=False, stable=True):
        if axis < 0:
            axis += x.ndim

        _len = ndx.asarray(x.shape[axis : axis + 1], dtype=dtypes.int64)._core()
        return _via_i64_f64(
            lambda x: opx.top_k(x, _len, largest=descending, axis=axis)[1], [x]
        )

    @validate_core
    def sort(self, x, *, axis=-1, descending=False, stable=True):
        if axis < 0:
            axis += x.ndim
        _len = ndx.asarray(x.shape[axis : axis + 1], dtype=dtypes.int64)._core()
        return _via_i64_f64(
            lambda x: opx.top_k(x, _len, largest=descending, axis=axis)[0], [x]
        )

    # statistical.py
    @validate_core
    def cumulative_sum(
        self,
        x,
        *,
        axis: int | None = None,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        include_initial: bool = False,
    ):
        if not isinstance(x.dtype, ndx.CoreType):
            return NotImplemented
        if axis is None:
            if x.ndim <= 1:
                axis = 0
            else:
                raise ValueError("axis must be specified for multi-dimensional arrays")
        out = from_corearray(
            opx.cumsum(
                x._core(), axis=opx.const(axis), exclusive=int(not include_initial)
            )
        )
        if dtype is not None:
            out = out.astype(dtype)
        return out

    @validate_core
    def max(self, x, *, axis=None, keepdims: bool = False):
        if axis is None:
            axes = []
        elif not isinstance(axis, Iterable):
            axes = [axis]
        else:
            axes = axis  # type: ignore

        if isinstance(x.dtype, dtypes.NullableFloating):
            fill_value = ndx.asarray(
                dtypes.get_finfo(x.dtype.values).min, x.dtype.values
            )
            x = ndx.where(x.null, fill_value, x.values)
        elif isinstance(x.dtype, dtypes.NullableIntegral):
            fill_value = ndx.asarray(
                dtypes.get_iinfo(x.dtype.values).min, x.dtype.values
            )
            x = ndx.where(x.null, fill_value, x.values)
        if not isinstance(x.dtype, dtypes.Numerical):
            raise TypeError("min is not supported for non-numeric types")

        return _via_i64_f64(
            lambda x: opx.reduce_max(
                x,
                opx.const(axes, dtype=dtypes.int64),
                keepdims=keepdims,
                noop_with_empty_axes=axis is not None,
            ),
            [x],
            cast_return=True,
        )

    @validate_core
    def mean(self, x, *, axis=None, keepdims: bool = False):
        return ndx.sum(x, axis=axis, keepdims=keepdims) / ndx.sum(
            ndx.full_like(x, 1, dtype=x.dtype), axis=axis, keepdims=keepdims
        )

    @validate_core
    def min(self, x, *, axis=None, keepdims: bool = False):
        if axis is None:
            axes = []
        elif not isinstance(axis, Iterable):
            axes = [axis]
        else:
            axes = axis  # type: ignore

        if isinstance(x.dtype, dtypes.NullableFloating):
            fill_value = ndx.asarray(
                dtypes.get_finfo(x.dtype.values).max, dtype=x.dtype.values
            )
            x = ndx.where(x.null, fill_value, x.values)
        elif isinstance(x.dtype, dtypes.NullableIntegral):
            fill_value = ndx.asarray(
                dtypes.get_iinfo(x.dtype.values).max, dtype=x.dtype.values
            )
            x = ndx.where(x.null, fill_value, x.values)
        if not isinstance(x.dtype, dtypes.Numerical):
            raise TypeError("min is not supported for non-numeric types")

        return _via_i64_f64(
            lambda x: opx.reduce_min(
                x,
                opx.const(axes, dtype=dtypes.int64),
                keepdims=keepdims,
                noop_with_empty_axes=axis is not None,
            ),
            [x],
            cast_return=True,
        )

    @validate_core
    def prod(
        self,
        x,
        *,
        axis=None,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        keepdims: bool = False,
    ):
        if axis is None:
            axes = []
        elif not isinstance(axis, Iterable):
            axes = [axis]
        else:
            axes = axis  # type: ignore

        x = x.astype(_determine_reduce_op_dtype(x, dtype, dtypes.uint32))

        if isinstance(x.dtype, dtypes.NullableNumerical):
            fill_value = ndx.asarray(1, dtype=x.dtype.values)
            x = ndx.where(x.null, fill_value, x.values)
        if not isinstance(x.dtype, dtypes.Numerical):
            raise TypeError("prod is not supported for non-numeric types")

        out = via_upcast(
            lambda x: opx.reduce_prod(
                x,
                opx.const(axes, dtype=dtypes.int64),
                keepdims=keepdims,
                noop_with_empty_axes=axis is not None,
            ),
            [x],
            int_dtype=dtypes.int64,
            float_dtype=dtypes.float32,
            cast_return=True,
        )

        return out

    @validate_core
    def clip(
        self,
        x,
        *,
        min=None,
        max=None,
    ):
        input = x.copy()
        if min is not None:
            min = ndx.asarray(min)
        if max is not None:
            max = ndx.asarray(max)
        if (
            min is not None
            and max is not None
            and min.ndim == 0
            and max.ndim == 0
            and isinstance(x.dtype, dtypes.Numerical)
        ):
            x, min, max = promote(x, min, max)
            if isinstance(x.dtype, dtypes._NullableCore):
                out_null = x.null
                x_values = x.values._core()
                clipped = from_corearray(opx.clip(x_values, min._core(), max._core()))
                clipped = Array._from_fields(x.dtype, values=clipped, null=out_null)
            else:
                clipped = from_corearray(opx.clip(x._core(), min._core(), max._core()))
            if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
                return ndx.where(ndx.isnan(input), np.nan, clipped)
            else:
                return clipped
        else:
            if max is not None:
                x = ndx.where(x <= max, x, max)
            if min is not None:
                x = ndx.where(x >= min, x, min)
            if isinstance(x.dtype, (dtypes.Floating, dtypes.NullableFloating)):
                return ndx.where(ndx.isnan(input), np.nan, x)
            else:
                return x

    @validate_core
    def std(
        self,
        x,
        axis=None,
        keepdims: bool = False,
        correction=0.0,
        dtype: dtypes.StructType | dtypes.CoreType | None = None,
    ):
        dtype = x.dtype if dtype is None else dtype
        if not isinstance(dtype, dtypes.CoreType):
            raise TypeError("std is not supported for non-core types")
        return ndx.sqrt(
            ndx.var(x, axis=axis, keepdims=keepdims, correction=correction, dtype=dtype)
        )

    @validate_core
    def sum(
        self,
        x,
        *,
        axis=None,
        dtype: dtypes.StructType | dtypes.CoreType | None = None,
        keepdims: bool = False,
    ):
        if axis is None:
            axes = []
        elif not isinstance(axis, Iterable):
            axes = [axis]
        else:
            axes = axis  # type: ignore

        x = x.astype(_determine_reduce_op_dtype(x, dtype, dtypes.uint64))

        # Fill any nulls with 0
        if isinstance(x.dtype, dtypes.NullableNumerical):
            x = ndx.where(x.null, ndx.asarray(0, x.values.dtype), x.values)

        if not isinstance(x.dtype, dtypes.Numerical):
            raise TypeError("sum is not supported for non-core types")

        out = _via_i64_f64(
            lambda corearray: opx.reduce_sum(
                corearray,
                opx.const(axes, dtype=dtypes.int64),
                keepdims=keepdims,
                noop_with_empty_axes=axis is not None,
            ),
            [x],
            cast_return=True,
        )

        return out

    @validate_core
    def var(
        self,
        x,
        *,
        axis=None,
        keepdims: bool = False,
        correction=0.0,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
    ):
        dtype = x.dtype if dtype is None else dtype
        if not isinstance(dtype, dtypes.CoreType):
            raise TypeError("var is not supported for non-core types")
        # We keepdims, so it is still broadcastable to x
        mean_ = self.mean(x, axis=axis, keepdims=True).astype(dtype)

        return self.sum(self.square(x - mean_), axis=axis, keepdims=keepdims).astype(
            dtype
        ) / (
            self.sum(ndx.full_like(x, 1), axis=axis, keepdims=keepdims).astype(dtype)
            - correction
        )

    @validate_core
    def make_nullable(self, x, null):
        if null.dtype != dtypes.bool:
            raise TypeError("'null' must be a boolean array")

        return ndx.Array._from_fields(
            dtypes.into_nullable(x.dtype),
            values=x.copy(),
            null=ndx.reshape(null, x.shape),
        )

    @validate_core
    def can_cast(self, from_, to) -> bool:
        if isinstance(from_, dtypes.CoreType) and isinstance(to, ndx.CoreType):
            return np.can_cast(from_.to_numpy_dtype(), to.to_numpy_dtype())
        return NotImplemented

    @validate_core
    def all(self, x, *, axis=None, keepdims: bool = False):
        if isinstance(x.dtype, dtypes._NullableCore):
            x = ndx.where(x.null, True, x.values)
        if functools.reduce(operator.mul, x._static_shape, 1) == 0:
            return ndx.asarray(True, dtype=ndx.bool)
        return ndx.min(x.astype(ndx.int8), axis=axis, keepdims=keepdims).astype(
            ndx.bool
        )

    @validate_core
    def any(self, x, *, axis=None, keepdims: bool = False):
        if isinstance(x.dtype, dtypes._NullableCore):
            x = ndx.where(x.null, False, x.values)
        if functools.reduce(operator.mul, x._static_shape, 1) == 0:
            return ndx.asarray(False, dtype=ndx.bool)
        return ndx.max(x.astype(ndx.int8), axis=axis, keepdims=keepdims).astype(
            ndx.bool
        )

    @validate_core
    def arange(self, start, stop=None, step=None, dtype=None, device=None) -> ndx.Array:
        step = ndx.asarray(step)
        if stop is None:
            start = ndx.asarray(start)
            stop = start
            start = ndx.asarray(0)
        elif start is None:
            start = ndx.asarray(0)

        start, stop, step = promote(start, stop, step)

        if not builtins.all(
            isinstance(x.dtype, ndx.CoreType) for x in (start, stop, step)
        ):
            raise ndx.UnsupportedOperationError(
                "Cannot perform arange with nullable `start`, `stop` or `step` parameters."
            )
        return ndx.astype(
            from_corearray(opx.range(start._core(), stop._core(), step._core())), dtype
        )

    @validate_core
    def tril(self, x, k=0) -> ndx.Array:
        if isinstance(x.dtype, dtypes._NullableCore):
            # NumPy appears to just ignore the mask so we do the same
            x = x.values
        return x._transmute(
            lambda core: opx.trilu(
                core, k=ndx.asarray(k, dtype=dtypes.int64)._core(), upper=0
            )
        )

    @validate_core
    def triu(self, x, k=0) -> ndx.Array:
        if isinstance(x.dtype, dtypes._NullableCore):
            # NumPy appears to just ignore the mask so we do the same
            x = x.values
        return x._transmute(
            lambda core: opx.trilu(
                core, k=ndx.asarray(k, dtype=dtypes.int64)._core(), upper=1
            )
        )

    @validate_core
    def eye(self, n_rows, n_cols=None, k=0, dtype=None, device=None) -> ndx.Array:
        if n_cols is None:
            n_cols = n_rows
        n_rows = ndx.asarray(n_rows, dtype=dtypes.int64)
        n_cols = ndx.asarray(n_cols, dtype=dtypes.int64)
        n_rows = ndx.reshape(n_rows, [-1])
        n_cols = ndx.reshape(n_cols, [-1])
        shape = ndx.concat([n_rows, n_cols])
        dummy = ndx.zeros(shape, dtype=dtype)
        return from_corearray(opx.eye_like(dummy._core(), k=k))

    @validate_core
    def linspace(
        self, start, stop, num, *, dtype=None, device=None, endpoint=True
    ) -> ndx.Array:
        return ndx.asarray(
            np.linspace(start, stop, num=num, endpoint=endpoint), dtype=dtype
        )

    @validate_core
    def ones(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        dtype = dtypes.float64 if dtype is None else dtype
        return ndx.full(shape, 1, dtype=dtype)

    @validate_core
    def ones_like(
        self, x, dtype: dtypes.StructType | dtypes.CoreType | None = None, device=None
    ):
        return ndx.full_like(x, 1, dtype=dtype)

    @validate_core
    def zeros(
        self,
        shape,
        dtype: dtypes.CoreType | dtypes.StructType | None = None,
        device=None,
    ):
        dtype = dtypes.float64 if dtype is None else dtype
        return ndx.full(shape, 0, dtype=dtype)

    @validate_core
    def zeros_like(
        self, x, dtype: dtypes.CoreType | dtypes.StructType | None = None, device=None
    ):
        return ndx.full_like(x, 0, dtype=dtype)

    @validate_core
    def empty(self, shape, dtype=None, device=None) -> ndx.Array:
        shape = ndx.asarray(shape, dtype=dtypes.int64)
        return ndx.full(shape, 0, dtype=dtype)

    @validate_core
    def empty_like(self, x, dtype=None, device=None) -> ndx.Array:
        return ndx.full_like(x, 0, dtype=dtype)


class NullableNumericOperationsImpl(NumericOperationsImpl, NullableOperationsImpl):
    def make_nullable(self, x, null):
        return NotImplemented


def _via_i64_f64(
    fn: Callable[..., _CoreArray], arrays: list[Array], *, cast_return=True
) -> ndx.Array:
    """Like ``_via_dtype`` but uses ``i64`` for integer or boolean types and ``float64``
    for floating point types.

    Raises TypeError if the provided arrays are neither.
    """
    return via_upcast(
        fn,
        arrays,
        int_dtype=dtypes.int64,
        float_dtype=dtypes.float64,
        use_unsafe_uint_cast=True,  # TODO this can cause overflow, we should set it to false and fix all uses
        cast_return=cast_return,
    )


def _determine_reduce_op_dtype(
    x: Array,
    dtype: dtypes.CoreType | dtypes.StructType | None,
    maximum_unsigned_dtype: dtypes.CoreType,
) -> dtypes.CoreType | dtypes.StructType:
    if dtype is not None:
        return dtype
    elif isinstance(x.dtype, dtypes.Unsigned):
        if ndx.iinfo(x.dtype).bits <= ndx.iinfo(maximum_unsigned_dtype).bits:
            return maximum_unsigned_dtype
        else:
            raise TypeError(f"Cannot reduce {x.dtype} to a smaller unsigned dtype")
    elif isinstance(x.dtype, dtypes.NullableUnsigned):
        if ndx.iinfo(x.dtype).bits <= ndx.iinfo(maximum_unsigned_dtype).bits:
            return dtypes.promote_nullable(maximum_unsigned_dtype)
        else:
            raise TypeError(f"Cannot reduce {x.dtype} to a smaller unsigned dtype")
    elif isinstance(x.dtype, dtypes.Integral):
        return dtypes.int64
    elif isinstance(x.dtype, dtypes.NullableIntegral):
        return dtypes.nint64
    else:
        return x.dtype



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING
from warnings import warn

import numpy as np
import numpy.typing as npt
from spox import Var

# FIXME: Remove private import from Spox! Better to just use reshape!
from spox._internal_op import unsafe_reshape

import ndonnx as ndx

if TYPE_CHECKING:
    from ndonnx import Array

    from ._corearray import _CoreArray


def promote(*args: Array | npt.ArrayLike | None) -> list[Array]:
    """Promote arguments following numpy's promotion rules.

    Constant scalars are converted to `Array` objects.

    `None` values are passed through.
    """
    arrays: list[Array] = []

    for arg in args:
        if isinstance(arg, ndx.Array):
            arrays.append(arg)
        elif isinstance(arg, (np.ndarray, np.generic)):
            arrays.append(ndx.asarray(arg))
        elif isinstance(arg, (float, int, str)):
            continue
        else:
            raise TypeError(f"Cannot promote {type(arg)}")

    if not arrays:
        raise ValueError("At least one array must be provided for type promotion")

    target_dtype = ndx.result_type(*arrays)
    string_dtypes = (ndx.utf8, ndx.nutf8)
    out: list[Array] = []
    for arg in args:
        # Deal with cross-kind scalar promotions
        if isinstance(arg, float) and not isinstance(
            target_dtype, (ndx.Floating, ndx.NullableFloating)
        ):
            target_dtype = ndx.result_type(target_dtype, ndx.float64)
        elif isinstance(arg, bool) and not target_dtype not in (ndx.bool, ndx.nbool):
            target_dtype = ndx.result_type(target_dtype, ndx.bool)
        elif isinstance(arg, int) and not isinstance(
            target_dtype, (ndx.Numerical, ndx.NullableNumerical)
        ):
            target_dtype = ndx.result_type(target_dtype, ndx.int64)

        if not isinstance(arg, ndx.Array):
            arg = ndx.asarray(np.array(arg))
        if arg.dtype in string_dtypes and target_dtype not in string_dtypes:
            raise TypeError("Cannot promote string scalar to numerical type")
        elif arg.dtype not in string_dtypes and target_dtype in string_dtypes:
            raise TypeError("Cannot promote non string scalar to string type")
        out.append(arg)

    return [arr.astype(target_dtype) for arr in out]


# We assume that rank will be static, because
# it is part if the type we are working on
def get_rank(tensor: _CoreArray | Var) -> int:
    return len(get_shape(tensor))


# We assume that type will be static, because
# it is part if the type we are working on
def get_dtype(tensor: _CoreArray | Var) -> np.dtype:
    return np.dtype(unwrap_var(tensor).unwrap_tensor().dtype)


def get_shape(x: _CoreArray | Var) -> tuple[int | str | None, ...]:
    if (shape := unwrap_var(x).unwrap_tensor().shape) is None:
        raise ValueError("Shape is not known")
    return shape


def set_shape(tensor: _CoreArray | Var, shape):
    tensor = unwrap_var(tensor)
    return unsafe_reshape(tensor, shape)


def unwrap_var(tensor: _CoreArray | Var) -> Var:
    if isinstance(tensor, Var):
        return tensor
    else:
        return tensor.var


def deprecated(msg: str):
    """Decorates a function as deprecated and raises a warning when it is called."""

    def _deprecated(fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            warn(msg, DeprecationWarning)
            return fn(*args, **kwargs)

        return inner

    return _deprecated



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import typing
from collections.abc import Callable
from typing import Union

import numpy as np
import spox.opset.ai.onnx.v19 as op
from spox import Tensor, Var, argument
from typing_extensions import Self

import ndonnx as ndx
import ndonnx._data_types as dtypes
from ndonnx.additional import shape
from ndonnx.additional._additional import _static_shape as static_shape

from ._corearray import _CoreArray
from ._index import ScalarIndexType
from ._utility import get_shape

if typing.TYPE_CHECKING:
    from ndonnx._data_types.coretype import CoreType
    from ndonnx._data_types.structtype import StructType

IndexType = Union[ScalarIndexType, tuple[ScalarIndexType, ...], "Array"]


def array(
    *,
    shape: tuple[int | str | None, ...],
    dtype: CoreType | StructType,
) -> Array:
    """Creates a new lazy ndonnx array. This is used to define inputs to an ONNX model.

    Parameters
    ----------
    shape : tuple[int | str | None, ...]
        The shape of the array. str dimensions denote symbolic dimensions and must be globally consistent. None dimensions denote unknown dimensions.
    dtype : CoreType | StructType
        The data type of the array.

    Returns
    -------
    out : Array
        The new array. This represents an ONNX model input.
    """
    return Array._construct(shape=shape, dtype=dtype)


def from_spox_var(
    var: Var,
) -> Array:
    """Construct an ``Array`` from a ``spox.Var``. This function is useful when
    interoperating between ``ndonnx`` and ``spox``.

    Parameters
    ----------
    var : spox.Var
        The variable to construct the array from.

    Returns
    -------
    out : Array
        The constructed array.
    """
    corearray = _CoreArray(var)
    return _from_corearray(corearray)


def _from_corearray(
    corearray: _CoreArray,
) -> Array:
    return Array._from_fields(corearray.dtype, data=corearray)


class Array:
    dtype: dtypes.StructType | dtypes.CoreType
    _fields: dict[str, Array | _CoreArray]

    def __init__(self, *args, **kwargs) -> None:
        raise TypeError("Array cannot be instantiated directly.")

    @classmethod
    def _from_fields(
        cls, dtype: StructType | CoreType, **fields: Array | _CoreArray
    ) -> Array:
        instance = cls.__new__(cls)
        instance.dtype = dtype
        field_dtypes = {name: field.dtype for name, field in fields.items()}
        expected_dtypes = (
            {"data": dtype} if isinstance(dtype, dtypes.CoreType) else dtype._fields()
        )
        if field_dtypes != expected_dtypes:
            raise TypeError(
                f"Expected fields `{expected_dtypes}` for `{dtype}`, received `{fields}`"
            )
        instance._fields = fields
        for name, field in fields.items():
            setattr(instance, name, field)
        return instance

    def __getattr__(self, name: str) -> Array:
        field = self._fields.get(name, None)
        if isinstance(field, Array):
            return field
        else:
            raise AttributeError(f"Field {name} not found")

    def _set(self, other: Array) -> Array:
        self.dtype = other.dtype
        self._fields = other._fields
        for name, field in other._fields.items():
            setattr(self, name, self._fields[name]._set(field))  # type: ignore
        return self

    def copy(self) -> Array:
        """Clones the array, copying all the array fields.

        Returns
        -------
        out : Array
            The cloned array.
        """
        return Array._from_fields(
            self.dtype,
            **{name: field.copy() for name, field in self._fields.items()},
        )

    def to_numpy(self) -> np.ndarray | None:
        """Returns the value of the Array as a NumPy array if available, otherwise
        ``None``."""
        field_values = {}
        for name, field in self._fields.items():
            if (eager_value := field.to_numpy()) is None:
                return None
            field_values[name] = eager_value
        return self.dtype._assemble_output(field_values)

    def astype(self, to: CoreType | StructType) -> Array:
        return ndx.astype(self, to)

    def __getitem__(self, index: IndexType) -> Array:
        if isinstance(index, Array) and not (
            isinstance(index.dtype, dtypes.Integral) or index.dtype == dtypes.bool
        ):
            raise TypeError(
                f"Index must be an integral or boolean 'Array', not `{index.dtype}`"
            )

        if isinstance(index, Array):
            index = index._core()

        return self._transmute(lambda corearray: corearray[index])

    def __setitem__(
        self, index: IndexType | Self, updates: int | bool | float | Array
    ) -> Array:
        if (
            isinstance(index, Array)
            and not isinstance(index.dtype, dtypes.Integral)
            and index.dtype != dtypes.bool
        ):
            raise TypeError(
                f"Index must be an integral or boolean 'Array', not `{index.dtype}`"
            )

        updates = ndx.asarray(updates).astype(self.dtype)

        for name, field in self._fields.items():
            if isinstance(field, _CoreArray):
                field[index._core() if isinstance(index, Array) else index] = (
                    updates._core()
                )
            else:
                field[index] = getattr(updates, name)

        return self

    def __repr__(self) -> str:
        eager_value = self.to_numpy()
        prefix = (
            f"Array({eager_value}, dtype={self.dtype})"
            if eager_value is not None
            else f"Array(dtype={self.dtype})"
        )
        return f"{prefix}"

    @staticmethod
    def _construct(
        shape: tuple[int | str | None, ...],
        dtype: CoreType | StructType,
        eager_values: dict | None = None,
    ) -> Array:
        if isinstance(dtype, dtypes.CoreType):
            var = (
                op.const(eager_values["data"], dtype=dtype.to_numpy_dtype())
                if eager_values is not None
                else argument(Tensor(dtype.to_numpy_dtype(), shape))
            )
            return Array._from_fields(
                dtype,
                data=_CoreArray(
                    eager_values["data"] if eager_values is not None else var
                ),
            )
        else:
            fields = {}
            for name, field_dtype in dtype._fields().items():
                fields[name] = Array._construct(
                    shape,
                    field_dtype,
                    eager_values[name] if eager_values is not None else None,
                )
            return Array._from_fields(
                dtype,
                **fields,
            )

    @property
    def ndim(self) -> int:
        """The rank of the array.

        This can be statically determined even for lazy arrays.
        """
        return len(self._static_shape)

    @property
    def shape(self) -> tuple[int | None, ...]:
        """The shape of the array.

        Returns:
            tuple[int | None, ...]: The shape of the array.

        Note that the shape of the array is a tuple of integers when the "eager value"
        of the array can be determined.

        When the array is lazy, getting the concrete shape may not be possible.
        We fall back to the `static_shape` function, which may be implemented
        using ONNX shape inference. This may have dimensions set as `None` where
        they are unknown or symbolic with respect to input shapes.
        """
        shape_array = shape(self).to_numpy()
        if shape_array is not None:
            return tuple(map(int, shape_array))
        else:
            return static_shape(self)

    @property
    def values(self) -> Array:
        """Accessor for data in a ``Array`` with nullable datatype."""
        if isinstance(self.dtype, dtypes.Nullable):
            return self.__getattr__("values")
        else:
            raise AttributeError("Field 'values' does not exist")

    @values.setter
    def values(self, value: Array) -> None:
        if (
            isinstance(self.dtype, dtypes.Nullable)
            and value.dtype == self.dtype._fields()["values"]
        ):
            self._fields["values"] = value
        else:
            raise ValueError("Field 'values' does not exist")

    @property
    def null(self) -> Array:
        """Accessor for missing null mask in a ``Array`` with nullable datatype."""
        if isinstance(self.dtype, dtypes.Nullable):
            return self.__getattr__("null")
        else:
            raise AttributeError("Field 'null' does not exist")

    @null.setter
    def null(self, value: Array) -> None:
        if isinstance(self.dtype, dtypes.Nullable) and value.dtype == dtypes.bool:
            self._fields["null"] = value
        else:
            raise ValueError("Field 'null' does not exist")

    @property
    def _static_shape(self) -> tuple[int | str | None, ...]:
        """Inferred shape of the array based on static ONNX shape propagation."""
        value = self.to_numpy()
        if value is not None:
            return value.shape
        else:
            current: _CoreArray | Array = self
            while not isinstance(current, _CoreArray):
                current = next(iter(current._fields.values()))
            return get_shape(current)

    def _transmute(self, op: Callable[[_CoreArray], _CoreArray]) -> Array:
        """Apply a layout transformation to the array (transmute the array).

        Data types must be preserved. This function will validate this and throw a ValueError if `op` breaks this contract.
        """
        fields: dict[str, _CoreArray | Array] = {}
        for name, field in self._fields.items():
            if isinstance(field, _CoreArray):
                fields[name] = op(field)
            else:
                fields[name] = field._transmute(op)
            if fields[name].dtype != field.dtype:
                raise ValueError(
                    f"Transmute operation changed the dtype of field {name} from {field.dtype} to {fields[name].dtype}. Only layout transformations are permitted."
                )
        return Array._from_fields(self.dtype, **fields)

    def _core(self) -> _CoreArray:
        if isinstance(self.dtype, dtypes.CoreType):
            return typing.cast(_CoreArray, self._fields["data"])
        raise ValueError("Only CoreType arrays have a core array")

    def spox_var(self) -> Var:
        """Return the underlying Spox ``Var``. This is useful when interoperating
        between ``Spox`` and ``ndonnx``.

        Raises
        ------
        TypeError
            If the data type of the ``Array`` is not a ``CoreType``, there is not a singular ``Var`` associated with it.
        """
        return self._core().var

    @staticmethod
    def __array_namespace__(*, api_version: str | None = None):
        if api_version is None:
            api_version = "2023.12"

        if api_version == "2023.12":
            return ndx
        else:
            raise ValueError(f"Unsupported API version {api_version}")

    def __float__(self) -> float:
        eager_value = self.to_numpy()
        if eager_value is not None and eager_value.size == 1:
            return float(eager_value)
        else:
            raise ValueError(f"Cannot convert Array of shape {self.shape} to a float")

    def __index__(self) -> int:
        eager_value = self.to_numpy()
        if (
            self.ndim == 0
            and eager_value is not None
            and isinstance(eager_value.item(), int)
        ):
            return int(eager_value)
        else:
            raise ValueError(f"Cannot convert Array {self} to an index (int)")

    def __int__(self) -> int:
        eager_value = self.to_numpy()
        if eager_value is None:
            raise ValueError(
                f"The lazy Array {self} cannot be converted to a scalar int"
            )
        if eager_value.size != 1:
            raise ValueError(
                f"Array {self} with more than one element cannot be converted to a scalar int"
            )
        return int(eager_value)

    def __bool__(self) -> bool:
        eager_value = self.to_numpy()
        if eager_value is None:
            raise ValueError(f"The truth value of lazy Array {self} is unknown")
        if eager_value.size != 1:
            raise ValueError(
                f"The truth value of Array {self} with more than one element is ambiguous"
            )
        return bool(eager_value)

    def __len__(self) -> int:
        if isinstance(self.shape[0], int):
            return self.shape[0]
        else:
            raise ValueError(f"Cannot determine length of Array {self}")

    def __pos__(self):
        return ndx.positive(self)

    def __neg__(self):
        return ndx.negative(self)

    def __abs__(self):
        return ndx.abs(self)

    def __add__(self, other):
        return ndx.add(self, other)

    def __radd__(self, other):
        return ndx.add(other, self)

    def __iadd__(self, other):
        return self._set(ndx.add(self, other))

    def __sub__(self, other):
        return ndx.subtract(self, other)

    def __rsub__(self, other):
        return ndx.subtract(other, self)

    def __isub__(self, other):
        return self._set(ndx.subtract(self, other))

    def __mul__(self, other):
        return ndx.multiply(self, other)

    def __rmul__(self, other):
        return ndx.multiply(other, self)

    def __imul__(self, other):
        return self._set(ndx.multiply(self, other))

    def __truediv__(self, other):
        return ndx.divide(self, other)

    def __rtruediv__(self, other):
        return ndx.divide(other, self)

    def __floordiv__(self, other):
        return ndx.floor_divide(self, other)

    def __rfloordiv__(self, other):
        return ndx.floor_divide(other, self)

    def __mod__(self, other):
        return ndx.remainder(self, other)

    def __rmod__(self, other):
        return ndx.remainder(other, self)

    def __pow__(self, other):
        return ndx.pow(self, other)

    def __rpow__(self, other):
        return ndx.pow(other, self)

    def __matmul__(self, other):
        return ndx.matmul(self, other)

    def __lt__(self, other):
        return ndx.less(self, other)

    def __le__(self, other):
        return ndx.less_equal(self, other)

    def __gt__(self, other):
        return ndx.greater(self, other)

    def __ge__(self, other):
        return ndx.greater_equal(self, other)

    def __invert__(self):
        return ndx.bitwise_invert(self)

    def __and__(self, other):
        return ndx.bitwise_and(self, other)

    def __or__(self, other):
        return ndx.bitwise_or(self, other)

    def __xor__(self, other):
        return ndx.bitwise_xor(self, other)

    def __lshift__(self, other):
        return ndx.bitwise_left_shift(self, other)

    def __rshift__(self, other):
        return ndx.bitwise_right_shift(self, other)

    def __eq__(self, other):
        if isinstance(other, type(Ellipsis)):
            # FIXME: What is going on here? Deserves a comment, IMHO
            return ndx.asarray(False)
        return ndx.equal(self, other)

    def __ne__(self, other):
        return ndx.not_equal(self, other)

    @property
    def mT(self) -> ndx.Array:  # noqa: N802
        """Transpose of a matrix (or a stack of matrices).

        If an array instance has fewer than two dimensions, an error should be raised.

        Returns
        -------
        out: Array
            array whose last two dimensions (axes) are permuted in reverse order relative to original array (i.e., for an array instance having shape ``(..., M, N)``, the returned array must have shape ``(..., N, M)``). The returned array must have the same data type as the original array.
        """
        return ndx.matrix_transpose(self)

    @property
    def size(self) -> ndx.Array:
        """Number of elements in the array.

        Returns
        -------
        out: Array
            Scalar ``Array`` instance whose value is the number of elements in the original array.
        """
        return ndx.prod(self.shape)

    @property
    def T(self) -> ndx.Array:  # noqa: N802
        """Transpose of the array.

        The array instance must be two-dimensional.

        Raises
        ------
        TypeError
            If the array instance is not two-dimensional.

        Returns
        -------
        out: Array
            Two-dimensional array whose first and last dimensions (axes) are permuted in reverse order relative to original array. The returned array must have the same data type as the original array.
        """
        if self.ndim != 2:
            raise TypeError("Cannot transpose Array of rank != 2")

        return ndx.matrix_transpose(self)

    def len(self):
        """Returns the length of the array as an array.

        The array instance must be one-dimensional.

        Raises
        ------
        TypeError
            If the array instance is not one-dimensional.

        Returns
        -------
        out: Array
            Singleton array whose value is the length of the original array.
        """
        if self.ndim != 1:
            raise TypeError("Cannot call len on Array of rank != 1")
        return shape(self)[0]

    def sum(self, axis: int | None = 0, keepdims: bool | None = False) -> ndx.Array:
        """See :py:func:`ndonnx.sum` for documentation."""
        return ndx.sum(self, axis=axis, keepdims=False)

    def prod(self, axis: int | None = 0, keepdims: bool | None = False) -> ndx.Array:
        """See :py:func:`ndonnx.prod` for documentation."""
        return ndx.prod(self, axis=axis, keepdims=False)

    def max(self, axis: int | None = 0, keepdims: bool | None = False) -> ndx.Array:
        """See :py:func:`ndonnx.max` for documentation."""
        return ndx.max(self, axis=axis, keepdims=False)

    def min(self, axis: int | None = 0, keepdims: bool | None = False) -> ndx.Array:
        """See :py:func:`ndonnx.min` for documentation."""
        return ndx.min(self, axis=axis, keepdims=False)

    def all(self, axis: int | None = 0, keepdims: bool | None = False) -> ndx.Array:
        """See :py:func:`ndonnx.all` for documentation."""
        return ndx.all(self, axis=axis, keepdims=False)

    def any(self, axis: int | None = 0, keepdims: bool | None = False) -> ndx.Array:
        """See :py:func:`ndonnx.any` for documentation."""
        return ndx.any(self, axis=axis, keepdims=False)


__all__ = [
    "Array",
    "array",
]



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

import ndonnx as ndx

from ._interface import OperationsBlock
from ._utils import validate_core


class NullableOperationsImpl(OperationsBlock):
    @validate_core
    def fill_null(self, x, value):
        value = ndx.asarray(value)
        if value.dtype != x.values.dtype:
            value = value.astype(x.values.dtype)
        return ndx.where(x.null, value, x.values)



# Copyright (c) QuantCo 2023-2024
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import warnings
from collections import namedtuple
from pathlib import Path

import numpy as np
import spox
import spox.opset.ai.onnx.v19 as op

import ndonnx as ndx

from ._corearray import _CoreArray

Arguments = namedtuple("Arguments", ["args", "kwargs"])
LazyEagerPair = namedtuple("LazyEagerPair", ["lazy", "value"])

try:
    import onnxruntime as ort

    ORT_PRESENT = True
except ImportError:
    ORT_PRESENT = False
    warnings.warn(
        "onnxruntime not found. Eager propagation and constant folding optimizations will be disabled."
    )


def eager_propagate(fn):
    def wrapper(*args, **kwargs):
        return _eager_propagate_with_custom_operators(None)(fn)(*args, **kwargs)

    return wrapper


def _eager_propagate_with_custom_operators(custom_operators_path: Path | None):
    def decorator(fn):
        def wrapper(*args, **kwargs):
            constant_inputs, lazy_args, flattened_inference_inputs = (
                _aggregate_arguments(Arguments(args, kwargs))
            )
            outputs = fn(*args, **kwargs)
            if constant_inputs and ORT_PRESENT:
                is_sequence = isinstance(outputs, (list, tuple))
                prediction = _propagate_helper(
                    fn, lazy_args, flattened_inference_inputs, custom_operators_path
                )
                outputs = [outputs] if not is_sequence else outputs
                corearrays: list[_CoreArray] = []
                for output in outputs:
                    if isinstance(output, _CoreArray):
                        corearrays.append(output)
                    else:
                        _get_corearrays(output, corearrays)
                for corearray_out, value in zip(corearrays, prediction):
                    if corearray_out.dtype == ndx.utf8:
                        value = value.astype(np.str_)
                    corearray_out._eager_value = value
                    corearray_out.var = op.const(value)

                return outputs[0] if not is_sequence else outputs
            else:
                return outputs

        return wrapper

    return decorator


def _aggregate_arguments(
    args: Arguments,
) -> tuple[bool, Arguments, list[LazyEagerPair]]:
    """Takes a function as input (which may only consist of _CoreArrays and various
    structures of it) and also the arguments to the function.

    Inputs:
        args: Arguments
            The positional and keyword arguments to the function.
    Returns:
        constant_inputs: bool
            Whether all the inputs are constant.
        lazy_args: Arguments
            The arguments to the function, but with all the _CoreArrays replaced with lazy variants.
        flattened_inference_inputs: list[LazyEagerPair]
            The constituent input _CoreArrays alongside their eager values.
    """

    constant_inputs = True
    lazy_args: Arguments = Arguments([], {})
    flattened_inference_inputs: list[LazyEagerPair] = []

    def collect_lazy_arguments(obj):
        nonlocal constant_inputs
        if isinstance(obj, _CoreArray):
            if obj.to_numpy() is None:
                constant_inputs = False
                return obj
            else:
                constant_inputs &= obj.to_numpy() is not None
                if obj.dtype == ndx.utf8:
                    # Lazy variant due to onnxruntime bug
                    lazy = _CoreArray(
                        spox.argument(
                            spox.Tensor(
                                obj.dtype.to_numpy_dtype(),
                                obj.var.unwrap_tensor().shape,
                            )
                        ),
                    )
                    flattened_inference_inputs.append(
                        LazyEagerPair(lazy, obj.to_numpy())
                    )
                    return lazy
                else:
                    return obj
        elif isinstance(obj, ndx.Array):
            obj_value = obj.to_numpy()
            if obj_value is None:
                constant_inputs = False
                return obj
            else:
                if obj.dtype in (ndx.utf8, ndx.nutf8):
                    # Lazy variant due to onnxruntime bug
                    lazy = ndx.array(shape=obj._static_shape, dtype=obj.dtype)  # type: ignore

                    # disassemble the array into its core_arrays
                    _flatten(obj, lazy, flattened_inference_inputs)
                    return lazy
                else:
                    return ndx.asarray(obj_value, obj.dtype)
        elif isinstance(obj, (list, tuple)):
            return type(obj)(map(collect_lazy_arguments, obj))
        elif isinstance(obj, dict):
            return {key: collect_lazy_arguments(value) for key, value in obj.items()}
        elif isinstance(obj, slice):
            return slice(
                collect_lazy_arguments(obj.start),
                collect_lazy_arguments(obj.stop),
                collect_lazy_arguments(obj.step),
            )
        else:
            return obj

    for arg in args.args:
        lazy_args.args.append(collect_lazy_arguments(arg))
    for key, value in args.kwargs.items():
        lazy_args.kwargs[key] = collect_lazy_arguments(value)
    return constant_inputs, lazy_args, flattened_inference_inputs


def _flatten(eager_array, lazy_array, flattened_inference_inputs):
    if isinstance(eager_array, _CoreArray):
        flattened_inference_inputs.append(
            LazyEagerPair(lazy_array, eager_array.to_numpy())
        )
    else:
        for name in eager_array._fields:
            _flatten(
                eager_array._fields[name],
                lazy_array._fields[name],
                flattened_inference_inputs,
            )


def _get_corearrays(array: ndx.Array, core_arrays: list[_CoreArray]):
    for name, field in array._fields.items():
        if isinstance(field, _CoreArray):
            core_arrays.append(field)
        else:
            _get_corearrays(field, core_arrays)


def _propagate_helper(
    fn, lazy_args, flattened_inference_inputs, custom_operators_path: Path | None = None
):
    lazy_outputs = fn(*lazy_args.args, **lazy_args.kwargs)
    is_sequence = isinstance(lazy_outputs, (list, tuple))
    if not is_sequence:
        lazy_outputs = [lazy_outputs]

    # destructure lazy_args and lazy_outputs into sequences of _CoreArrays for ModelProto creation

    model_proto = ndx.build(
        {
            f"input_{i}": input.lazy
            for i, input in enumerate(flattened_inference_inputs)
        },
        {
            f"output_{i}": output_corearray
            for i, output_corearray in enumerate(lazy_outputs)
        },
    )

    sess_options = ort.SessionOptions()
    if custom_operators_path is not None:
        sess_options.register_custom_ops_library(str(custom_operators_path))
    sess = ort.InferenceSession(
        model_proto.SerializeToString(), sess_options=sess_options
    )
    prediction = sess.run(
        None,
        {
            f"input_{i}": input.value
            for i, input in enumerate(flattened_inference_inputs)
        },
    )

    return prediction
